﻿Imports System.Collections
''' <summary>
''' A collection of key-value pairs of a ini section.
''' </summary>
Public Class IniFileKeyValuePairCollection
    Inherits ObjectModel.Collection(Of IniFileKeyValuePair)

    Private iecHeaderComparer As IEqualityComparer(Of String)
    Private ifsSection As IniFileSection

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileKeyValuePairCollection class.
    ''' </summary>
    ''' <param name="section">The section whose key-value pairs are stored in this instance of the IniLib.IniFileKeyValuePairCollection class.</param>
    ''' <param name="keyComparer">The comparer to use when comparing two keys.</param>
    Public Sub New(ByVal section As IniFileSection, ByVal keyComparer As IEqualityComparer(Of String))
        iecHeaderComparer = keyComparer
        ifsSection = section
    End Sub

    ''' <summary>
    ''' Inserts a key-value pair at the specified index.
    ''' </summary>
    ''' <param name="index">The index at which to insert the key-value pair.</param>
    ''' <param name="item">The key-value pair to insert at the specified index.</param>
    Protected Overrides Sub InsertItem(ByVal index As Integer, ByVal item As IniFileKeyValuePair)
        If Not item.IsCommentOrEmpty Then
            If ifsSection.IsSectionHeaderVisible = False Then
                Throw New ArgumentException("Key-value pairs that are not empty or comments can not be added to sections whose section header is invisible.")
            End If
            If (ifsSection.File.Flags And IniFileFlags.SupportsKeyValuePairKeyRepetition) = IniFileFlags.Default AndAlso Me.Contains(item.Key) Then
                Throw New ArgumentException("Collection already contains an item with the specified key.")
            End If
        End If
        MyBase.InsertItem(index, item)
        item.SetParentSection(ifsSection)
    End Sub

    ''' <summary>
    ''' Replaces the key-value pair currently at the specified index by the specified item.
    ''' </summary>
    ''' <param name="index">The index at which to set the key-value pair.</param>
    ''' <param name="item">The key-value pair replacing the key-value pair which currently is at the index specified by index.</param>
    Protected Overrides Sub SetItem(ByVal index As Integer, ByVal item As IniFileKeyValuePair)
        If Not item.IsCommentOrEmpty AndAlso (ifsSection.File.Flags And IniFileFlags.SupportsKeyValuePairKeyRepetition) = IniFileFlags.Default AndAlso Me.Contains(item.Key) AndAlso Not iecHeaderComparer.Equals(item.Key, Me.Items(index).Key) Then
            Throw New ArgumentException("Collection already contains an item with the specified key.")
        End If
        Me.Item(index).ParentSection = Nothing
        MyBase.SetItem(index, item)
        item.SetParentSection(ifsSection)
    End Sub

    ''' <summary>
    ''' Removes the item which is at the specified index.
    ''' </summary>
    ''' <param name="index">The index of the item to remove.</param>
    ''' <remarks></remarks>
    Protected Overrides Sub RemoveItem(ByVal index As Integer)
        Dim itm As IniFileKeyValuePair = Item(index)
        MyBase.RemoveItem(index)
        itm.SetParentSection(Nothing)
    End Sub

    ''' <summary>
    ''' Removes all key-value pairs from the collection.
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub ClearItems()
        For i As Integer = Me.Items.Count - 1 To 0 Step -1
            Remove(Items(i))
        Next
    End Sub

    ''' <summary>
    ''' The parent section of all key-value pairs stored in this instance of IniLib.IniFileKeyValuePairCollection class.
    ''' </summary>
    Public ReadOnly Property Section As IniFileSection
        Get
            Return ifsSection
        End Get
    End Property

    ''' <summary>
    ''' Retrieves whether the key-value pair collection contains a key-value pair with the specified key.
    ''' </summary>
    ''' <param name="key">The key to search in the collection.</param>
    Public Overloads Function Contains(ByVal key As String) As Boolean
        For Each c As IniFileKeyValuePair In Me
            If iecHeaderComparer.Equals(c.Key, key) Then
                Return True
            End If
        Next
        Return False
    End Function

    ''' <summary>
    ''' Gets the index of the first key-value pair whose key equals the specified key.
    ''' </summary>
    ''' <param name="key">The key whose index to get.</param>
    ''' <returns>Returns either the index of the key-value pair with the specified key or -1 if there aren't any key-value pairs with the specified key.</returns>
    Public Overloads Function IndexOf(ByVal key As String) As Integer
        For i As Integer = 0 To Count - 1
            If iecHeaderComparer.Equals(Me.Item(i).Key, key) Then
                Return i
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Gets the index of the first key-value pair whose key equals the specified key starting at the specified start index.
    ''' </summary>
    ''' <param name="key">The key to look for.</param>
    ''' <param name="startIndex">The index at which to start searching for a key-value pair with the specified key.</param>
    ''' <returns>Returns either the index of the key-value pair with the specified key or -1 if there aren't any key-value pairs with the specified key.</returns>
    Public Overloads Function IndexOf(ByVal key As String, ByVal startIndex As Integer) As Integer
        For i As Integer = startIndex To Count - 1
            If iecHeaderComparer.Equals(Me.Item(i).Key, key) Then
                Return i
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Gets the first key-value pair with the specified key.
    ''' </summary>
    ''' <param name="key">The key of the key-value pair to look for.</param>
    ''' <value></value>
    ''' <returns>Returns either the first key-value pair with the specified key or null if there aren't any key-value pairs with the specified key.</returns>
    Default Public Overloads ReadOnly Property Item(ByVal key As String) As IniFileKeyValuePair
        Get
            For i As Integer = 0 To Count - 1
                If iecHeaderComparer.Equals(Me.Item(i).Key, key) Then
                    Return Item(i)
                End If
            Next
            Return Nothing
        End Get
    End Property

    ''' <summary>
    ''' Invokes the specified action for all key-value pairs in the collection whose key is equal to the specified key.
    ''' </summary>
    ''' <param name="key">The key of the key-value pairs to invoke the specified action for.</param>
    ''' <param name="action">The action to invoke for all key-value pairs with the specified key in the collection.</param>
    ''' <remarks></remarks>
    Public Sub ForEach(ByVal key As String, ByVal action As Action(Of IniFileKeyValuePair))
        For Each itm As IniFileKeyValuePair In Me
            If iecHeaderComparer.Equals(itm.Key, key) Then
                action(itm)
            End If
        Next
    End Sub

    ''' <summary>
    ''' Returns the comparer to use when comparing two key-value pairs' keys.
    ''' </summary>
    Public ReadOnly Property Comparer As IEqualityComparer(Of String)
        Get
            Return iecHeaderComparer
        End Get
    End Property
End Class
