﻿Public MustInherit Class IniTextWriter
    ''' <summary>
    ''' Writes the specified characters into a buffer.
    ''' </summary>
    ''' <param name="chars">Specifies the characters to write into a buffer.</param>
    Protected MustOverride Sub Write(ByVal ParamArray chars() As Char)

    ''' <summary>
    ''' Flushes the text writer.
    ''' </summary>
    Public Overridable Sub Flush()
    End Sub

    ''' <summary>
    ''' Writes the specified string into a buffer.
    ''' </summary>
    ''' <param name="text">Specifies the string to write into a buffer.</param>
    Protected Overridable Sub Write(ByVal text As String)
        Write(text.ToCharArray())
    End Sub

    ''' <summary>
    ''' Returns the character used to separate keys and values.
    ''' </summary>
    Public Overridable ReadOnly Property KeyValueSeparator As Char
        Get
            Return "="c
        End Get
    End Property

    ''' <summary>
    ''' Returns the character used to separate key-value pairs and comments.
    ''' </summary>
    Public Overridable ReadOnly Property CommentSeparator As Char
        Get
            Return ";"c
        End Get
    End Property

    ''' <summary>
    ''' Specifies whether empty values are valid.
    ''' </summary>
    Protected Overridable ReadOnly Property AllowEmptyValues As Boolean
        Get
            Return False
        End Get
    End Property

    ''' <summary>
    ''' Writes an ini file section into a buffer.
    ''' </summary>
    ''' <param name="section">Specifies the section to write into a buffer.</param>
    Public Sub WriteSection(ByVal section As IniFileSection)
        WriteSectionHeader(section)
        For Each kvp As IniKeyValuePairDescription In section.KeyValues
            WriteKeyValuePair(kvp)
        Next
    End Sub

    ''' <summary>
    ''' Writes an ini section into the buffer taking its section name and key-value pairs.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name of the section.</param>
    ''' <param name="keyValuePairs">Specifies the key-value pairs within the section.</param>
    Public Sub WriteSection(ByVal sectionName As String, ByVal keyValuePairs As IEnumerable(Of IniKeyValuePairDescription))
        If Not String.IsNullOrEmpty(sectionName) Then
            Write("["c)
            Write(sectionName)
            Write("]"c)
        End If
        For Each kvp As IniKeyValuePairDescription In keyValuePairs
            WriteKeyValuePair(kvp)
        Next
    End Sub

    ''' <summary>
    ''' Writes an ini section into the buffer taking its section name, section comment and key-value pairs.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name of the section.</param>
    ''' <param name="sectionComment">Specifies the section header comment.</param>
    ''' <param name="keyValuePairs">Specifies the key-value pairs within the section.</param>
    Public Sub WriteSection(ByVal sectionName As String, ByVal sectionComment As String, ByVal keyValuePairs As IEnumerable(Of IniKeyValuePairDescription))
        If Not String.IsNullOrEmpty(sectionName) Then
            Write("["c)
            Write(sectionName)
            Write("]"c)
        End If
        If sectionComment IsNot Nothing Then
            WriteComment(sectionComment)
        Else
            Write(Environment.NewLine)
        End If
        For Each kvp As IniKeyValuePairDescription In keyValuePairs
            WriteKeyValuePair(kvp)
        Next
    End Sub

    Public Sub WriteSectionHeader(ByVal section As IniFileSection)
        If section.IsSectionNameVisible Then
            Write("["c)
            Write(section.SectionName)
            Write("]"c)
        End If
        If section.Comment IsNot Nothing Then
            WriteComment(section.Comment)
        End If
        Write(Environment.NewLine)
    End Sub

    Public Sub WriteKeyValuePair(ByVal keyValuePair As IniKeyValuePairDescription)
        If Not keyValuePair.IsCommentOrEmpty Then
            WriteKeyFormatted(keyValuePair.Key)
            Write(KeyValueSeparator)
            WriteValueFormatted(keyValuePair.Value)
        End If
        If keyValuePair.Comment IsNot Nothing Then
            WriteComment(keyValuePair.Comment)
        End If
        Write(Environment.NewLine)
    End Sub

    ''' <summary>
    ''' Formats a key and writes it into a buffer.
    ''' </summary>
    ''' <param name="key">Specifies the key to write into a buffer.</param>
    Protected Overridable Sub WriteKeyFormatted(ByVal key As String)
        Write(key)
        Write(" "c)
    End Sub

    ''' <summary>
    ''' Formats a value and writes it into a buffer.
    ''' </summary>
    ''' <param name="value">Specifies the key to write into a buffer.</param>
    Protected Overridable Sub WriteValueFormatted(ByVal value As String)
        If value IsNot Nothing Then
            Write(" "c)
            Write(value)
        End If
    End Sub

    ''' <summary>
    ''' Writes a comment into a buffer.
    ''' </summary>
    ''' <param name="comment">Specifies the comment to write into a buffer.</param>
    ''' <remarks></remarks>
    Public Sub WriteComment(ByVal comment As String)
        Write(";"c)
        Write(comment)
    End Sub
End Class
