﻿Imports System.Collections
Imports System.Runtime.Serialization, Microsoft.VisualBasic

''' <summary>
''' Defines a serializer that can be used to serialize classes into an ini file.
''' </summary>
''' <remarks>The serialization of objects is based on the SerializableAttribute-Attribute. All objects that are not primitive or strings must be marked as serializable. Types that do not implement the ISerializable-Attribute are serialized in the following way:
''' All fields that are public and instance members are serialized if they are not marked as unserializable (by the System.NonSerializedAttribute attribute). All unindexed properties that are public and instance members are serialized if they are not marked as unserializable (by the System.NonSerializedAttribute attribute).
''' Types that implement the ISerializable interface must be serialized to the specified System.Runtime.Serialization.SerializationInfo. Moreover apart from the abstract methods of the ISerializable-interface there must be a constructor of the form .ctr(System.Runtime.Serialization.SerializationInfo, System.Runtime.Serialization.StreamingContext) that deserializes an object.</remarks>
Public Class IniFileSerializer
    Private chrSeparator As Char

    ''' <summary>
    ''' Initializes a new instance of IniFileSerializer type.
    ''' </summary>
    Public Sub New()
        chrSeparator = "."c
    End Sub

    ''' <summary>
    ''' Initializes a new instance of IniFileSerializer type.
    ''' </summary>
    ''' <param name="separator">Specifies the separator used for sub-elements.</param>
    Public Sub New(ByVal separator As Char)
        If Not Char.IsLetterOrDigit(separator) Then
            chrSeparator = separator
        Else
            Throw New ArgumentException("Separators mustnot be letters or digits.")
        End If
    End Sub

    ''' <summary>
    ''' Serializes the specified value using the specified object name as an identifier. The identifier must be unique.
    ''' </summary>
    ''' <param name="objectName">An unique identifier that represents the object.</param>
    ''' <param name="value">A value to serialize.</param>
    ''' <param name="file">Specifies the ini file to use for serialization.</param>
    ''' <remarks>Object names should be valid. Use IsValidObjectName(String) to check the validity of object names.</remarks>
    Public Sub Serialize(ByVal objectName As String, ByVal value As Object, ByVal file As IniFile)
        If IsValidObjectName(objectName, file) Then
            SerializeObject(file.CreateSection(objectName), value, New StreamingContext(StreamingContextStates.All), file)
        Else
            Throw New ArgumentException("The specified object name is not valid as it contains either the separator-char used by the serializer, the key-value separator or the comment separator used by the ini file.")
        End If
    End Sub

    ''' <summary>
    ''' Represents the core method to serialize objects. Override to provide extended serialization.
    ''' </summary>
    ''' <param name="targetSection">Represents the section the data describing the value is stored in.</param>
    ''' <param name="value">A value to serialize.</param>
    ''' <param name="context">The context under which data is serialized.</param>
    Protected Overridable Sub SerializeObject(ByVal targetSection As IniFileSection, ByVal value As Object, ByVal context As StreamingContext, ByVal file As IniFile)
        If value Is Nothing Then
            targetSection.CreateKeyValuePair(".type", GetType(Object).AssemblyQualifiedName)
            targetSection.CreateKeyValuePair(".value", "NULL")
        ElseIf value.GetType().IsPrimitive OrElse TypeOf value Is String Then
            targetSection.CreateKeyValuePair(".type", value.GetType().AssemblyQualifiedName)
            targetSection.CreateKeyValuePair(".value", value.ToString())
        ElseIf value.GetType().IsSerializable Then
            If TypeOf value Is ISerializable Then
                Dim info As New SerializationInfo(value.GetType(), New FormatterConverter())
                Dim subobjectName As String
                DirectCast(value, ISerializable).GetObjectData(info, context)
                targetSection.CreateKeyValuePair(".type", value.GetType().AssemblyQualifiedName)

                For Each se As SerializationEntry In info
                    If IsValidObjectName(se.Name, file) Then
                        subobjectName = String.Concat(targetSection.SectionName, chrSeparator, se.Name)
                        targetSection.CreateKeyValuePair(se.Name, subobjectName)
                        SerializeObject(file.CreateSection(subobjectName), se.Value, context, file)
                    Else
                        Throw New ArgumentException(String.Concat("The serialization entry's name is invalid for the serializer: '", se.Name, "'"))
                    End If
                Next
            ElseIf TypeOf value Is IEnumerable Then
                SerializeEnumerable(targetSection, DirectCast(value, IEnumerable), context)
            Else
                Dim subobjectName As String
                targetSection.CreateKeyValuePair(".type", value.GetType().AssemblyQualifiedName)
                For Each fi As Reflection.FieldInfo In value.GetType().GetFields(Reflection.BindingFlags.Public Or Reflection.BindingFlags.Instance)
                    If Not fi.IsInitOnly AndAlso Not fi.IsDefined(GetType(NonSerializedAttribute), False) Then
                        If fi.FieldType.IsDefined(GetType(SerializableAttribute), False) Then
                            If IsValidObjectName(fi.Name, file) Then
                                subobjectName = String.Concat(targetSection.SectionName, chrSeparator, fi.Name)
                                targetSection.CreateKeyValuePair(fi.Name, subobjectName)
                                SerializeObject(file.CreateSection(subobjectName), fi.GetValue(value), context, file)
                            Else
                                Throw New ArgumentException(String.Concat("The fields's name is not allowed by the serializer: '", fi.Name, "'"))
                            End If
                        Else
                            Throw New ArgumentException(String.Concat("Type ", fi.FieldType.FullName, " could not be serialized."))
                        End If
                    End If
                Next
                For Each pi As Reflection.PropertyInfo In value.GetType().GetProperties(Reflection.BindingFlags.Public Or Reflection.BindingFlags.Instance)
                    If pi.CanRead AndAlso pi.CanWrite AndAlso Not pi.IsDefined(GetType(NonSerializedAttribute), False) Then
                        If pi.PropertyType.IsDefined(GetType(SerializableAttribute), False) Then
                            If IsValidObjectName(pi.Name, file) Then
                                subobjectName = String.Concat(targetSection.SectionName, chrSeparator, pi.Name)
                                targetSection.CreateKeyValuePair(pi.Name, subobjectName)
                                SerializeObject(file.CreateSection(subobjectName), pi.GetValue(value, {}), context, file)
                            Else
                                Throw New ArgumentException(String.Concat("The fields's name is not allowed by the serializer: '", pi.Name, "'"))
                            End If
                        Else
                            Throw New ArgumentException(String.Concat("Type ", pi.PropertyType.FullName, " could not be serialized."))
                        End If
                    End If
                Next
            End If
        Else
            Throw New ArgumentException("The specified value could not be serialized.")
        End If
    End Sub

    ''' <summary>
    ''' Serializes classes implementing the IEnumerable-interface that do not implement ISerializable.
    ''' </summary>
    ''' <param name="targetSection">Represents the section the data describing the value is stored in.</param>
    ''' <param name="enumerable">An enumerator to serialize.</param>
    ''' <param name="context">The context under which data is serialized.</param>
    Protected Overridable Sub SerializeEnumerable(ByVal targetSection As IniFileSection, ByVal enumerable As IEnumerable, ByVal context As StreamingContext)
        Dim subobjectName As String
        Dim subsection As IniFileSection
        Dim currentKeyValuePair As IniFileKeyValuePair
        Dim itemName As String
        Dim index As Long = 0
        Dim lengthKeyValuePair As IniFileKeyValuePair = targetSection.CreateKeyValuePair(".totalLength")

        targetSection.CreateKeyValuePair(".type", enumerable.GetType().AssemblyQualifiedName)
        If enumerable.GetType().IsArray Then
            Dim sb As New System.Text.StringBuilder()
            sb.Append(DirectCast(enumerable, Array).GetLength(0).ToString())
            For i As Integer = 1 To DirectCast(enumerable, Array).Rank - 1
                sb.Append(", ")
                sb.Append(DirectCast(enumerable, Array).GetLongLength(i).ToString())
            Next
            targetSection.CreateKeyValuePair(".lengths", sb.ToString())
        End If

        For Each objectValue As Object In enumerable
            itemName = "item" & index.ToString()
            subobjectName = String.Concat(targetSection.SectionName, chrSeparator, itemName)
            currentKeyValuePair = targetSection.CreateKeyValuePair(itemName, subobjectName)
            subsection = targetSection.File.CreateSection(subobjectName)
            index += 1L
            SerializeObject(subsection, objectValue, context, targetSection.File)
        Next
        lengthKeyValuePair.Value = index.ToString()
    End Sub

    ''' <summary>
    ''' Gets the Add(object)-method of a type. The returned value is used to deserialize enumerations and can be extended or optimized by overriding it.
    ''' </summary>
    ''' <param name="type">Specifies the type whose Add(object)-Method to get.</param>
    Protected Overridable Function GetAddMethod(ByVal type As Type) As Reflection.MethodInfo
        Dim param As Reflection.ParameterInfo()
        For Each methods As Reflection.MethodInfo In type.GetMethods(Reflection.BindingFlags.Public Or Reflection.BindingFlags.Instance)
            If methods.Name = "Add" Then
                param = methods.GetParameters()
                If param.Length = 1 Then
                    If param(0).ParameterType Is type OrElse type.IsSubclassOf(param(0).ParameterType) Then
                        Return methods
                    End If
                End If
            End If
        Next
        For Each tp As Type In type.GetInterfaces()
            For Each m As Reflection.MethodInfo In tp.GetMethods(Reflection.BindingFlags.Public Or Reflection.BindingFlags.Instance)
                If m.Name = "Add" Then
                    param = m.GetParameters()
                    If param.Length = 1 Then
                        If param(0).ParameterType Is type OrElse type.IsSubclassOf(param(0).ParameterType) Then
                            Return m
                        End If
                    End If
                End If
            Next
        Next
        Throw New NotImplementedException("The specified type does not implement an Add(object) method.")
    End Function

    ''' <summary>
    ''' Deserializes a value implementing the IEnumerable-interface.
    ''' </summary>
    ''' <param name="sourceSection">Specifies the section whose content to load.</param>
    ''' <param name="context">The context under which data is deserialized.</param>
    Protected Overridable Function DeserializeEnumerable(ByVal sourceSection As IniFileSection, ByVal context As StreamingContext, ByVal file As IniFile) As Object
        Dim type As Type = Type.GetType(sourceSection.KeyValues(".type").Value)
        Dim totalLength As Long = Long.Parse(sourceSection.KeyValues(".totalLength").Value)
        If type.IsArray Then
            Dim ctr() As Reflection.ConstructorInfo = type.GetConstructors(Reflection.BindingFlags.Public Or Reflection.BindingFlags.NonPublic Or Reflection.BindingFlags.Instance)
            Dim lengths() As Long = Array.ConvertAll(Of String, Long)(sourceSection.KeyValues(".lengths").Value.Split(","c), Function(arg As String) Integer.Parse(arg.Trim()))
            Dim instance As Array = Array.CreateInstance(type.GetElementType(), lengths)
            If totalLength <= Integer.MaxValue Then
                Dim indices(lengths.Length - 1) As Integer
                For i As Integer = 0 To CInt(totalLength) - 1
                    instance.SetValue(DeserializeObject(file.Sections(sourceSection.KeyValues("item" & i.ToString()).Value), context, file), indices)
                    IncrementIndices(indices, lengths)
                Next
            Else
                Dim indices(lengths.Length - 1) As Long
                For l As Long = 0L To totalLength - 1L
                    instance.SetValue(DeserializeObject(file.Sections(sourceSection.KeyValues("item" & l.ToString()).Value), context, file), indices)
                    IncrementIndices(indices, lengths)
                Next
            End If
            Return instance
        Else
            Dim instance As Object = Activator.CreateInstance(type)
            Dim currentObject As Object
            If totalLength <= Integer.MaxValue Then
                For i As Integer = 0 To CInt(totalLength) - 1
                    currentObject = DeserializeObject(file.Sections(sourceSection.KeyValues("item" & i.ToString()).Value), context, file)
                    GetAddMethod(currentObject.GetType()).Invoke(instance, {currentObject})
                Next
            Else
                For l As Long = 0L To totalLength - 1L
                    currentObject = DeserializeObject(file.Sections(sourceSection.KeyValues("item" & l.ToString()).Value), context, file)
                    GetAddMethod(currentObject.GetType()).Invoke(instance, {currentObject})
                Next
            End If
            Return instance
        End If
    End Function

    'Increment indices for arrays
    Private Sub IncrementIndices(ByVal indices() As Integer, ByVal lengths() As Long)
        For i As Integer = indices.GetUpperBound(0) To indices.GetLowerBound(0) Step -1
            If indices(i) < lengths(i) - 1 Then
                indices(i) += 1
                Exit For
            End If
        Next
    End Sub

    Private Sub IncrementIndices(ByVal indices() As Long, ByVal lengths() As Long)
        For i As Integer = indices.GetUpperBound(0) To indices.GetLowerBound(0) Step -1
            If indices(i) < lengths(i) - 1L Then
                indices(i) += 1L
                Exit For
            End If
        Next
    End Sub

    ''' <summary>
    ''' Deserializes the specified value using the specified object name as an identifier. The identifier must be unique.
    ''' </summary>
    ''' <param name="objectName">An unique identifier that represents the object.</param>
    ''' <remarks>Object names should be valid. Use IsValidObjectName(String) to check the validity of object names.</remarks>
    Public Function Deserialize(ByVal objectName As String, ByVal file As IniFile) As Object
        If IsValidObjectName(objectName, file) Then
            Return DeserializeObject(file.Sections(objectName), New StreamingContext(StreamingContextStates.All), file)
        Else
            Throw New ArgumentException("The specified object name is not valid as it contains either the separator-char used by the serializer, the key-value separator or the comment separator used by the ini file.")
        End If
    End Function

    ''' <summary>
    ''' Represents the core method to deserialize objects. Override to provide extended serialization.
    ''' </summary>
    ''' <param name="section">Represents the section the data describing the value is stored in.</param>
    ''' <param name="context">The context under which data is deserialized.</param>
    Protected Overridable Function DeserializeObject(ByVal section As IniFileSection, ByVal context As StreamingContext, ByVal file As IniFile) As Object
        Dim type As Type = Type.GetType(section.KeyValues(".type").Value)
        If type Is GetType(Object) Then
            If StringComparer.InvariantCultureIgnoreCase.Equals(section.KeyValues(".value"), "NULL") Then
                Return Nothing
            Else
                Throw New InvalidCastException(String.Concat("Object could not be serialized due to an invalid entry. The following sequence can not be converted into System.Object: """, section.KeyValues(".value").Value, """."))
            End If
        ElseIf type.IsPrimitive OrElse type Is GetType(String) Then
            Return Convert.ChangeType(section.KeyValues(".value").Value, type)
        ElseIf type.IsSerializable Then
            If type.GetInterface(GetType(ISerializable).FullName) IsNot Nothing Then
                Dim ctr As Reflection.ConstructorInfo = type.GetConstructor(Reflection.BindingFlags.Public Or Reflection.BindingFlags.NonPublic Or Reflection.BindingFlags.Static Or Reflection.BindingFlags.Instance, Nothing, {GetType(SerializationInfo), GetType(StreamingContext)}, Nothing)
                If ctr IsNot Nothing Then
                    Dim serializationinfo As New SerializationInfo(type, New FormatterConverter())

                    For Each kvp As IniFileKeyValuePair In section.KeyValues
                        If kvp.Key(0) <> "."c Then
                            serializationinfo.AddValue(kvp.Key, DeserializeObject(file.Sections(kvp.Value), context, file))
                        End If
                    Next
                    Return ctr.Invoke({serializationinfo, context})
                Else
                    Throw New SerializationException(String.Concat("Constructor ", type.FullName, "(System.Runtime.Serialization.SerializationInfo, System.Runtime.Serialization.StreamingContext) missing."))
                End If
            ElseIf type.GetInterface(GetType(IEnumerator).FullName) IsNot Nothing Then
                Dim instance As Object = Activator.CreateInstance(type)
                Dim members As Reflection.MemberInfo()
                For Each kvp As IniFileKeyValuePair In section.KeyValues
                    If Not String.IsNullOrEmpty(kvp.Key) AndAlso kvp.Key(0) <> "."c Then
                        members = type.GetMember(kvp.Key, Reflection.BindingFlags.Public Or Reflection.BindingFlags.Instance)
                        If members.Length = 1 Then
                            If TypeOf members(0) Is Reflection.PropertyInfo Then
                                DirectCast(members(0), Reflection.PropertyInfo).SetValue(instance, DeserializeObject(file.Sections(kvp.Value), context, file), {})
                            ElseIf TypeOf members(0) Is Reflection.FieldInfo Then
                                DirectCast(members(0), Reflection.FieldInfo).SetValue(instance, DeserializeObject(file.Sections(kvp.Value), context, file))
                            End If
                        ElseIf members.Length > 1 Then
                            Dim param As Reflection.ParameterInfo()
                            For Each m As Reflection.MemberInfo In members
                                If TypeOf m Is Reflection.PropertyInfo Then
                                    With DirectCast(m, Reflection.PropertyInfo)
                                        param = .GetIndexParameters()
                                        If param.Length = 0 Then
                                            .SetValue(instance, DeserializeObject(file.Sections(kvp.Value), context, file), {})
                                        End If
                                    End With
                                    Exit For
                                ElseIf TypeOf m Is Reflection.FieldInfo Then
                                    DirectCast(m, Reflection.FieldInfo).SetValue(instance, DeserializeObject(file.Sections(kvp.Value), context, file))
                                    Exit For
                                End If
                            Next
                        End If
                    End If
                Next
                Return instance
            Else
                Return DeserializeEnumerable(section, context, file)
            End If
        Else
            Throw New ArgumentException(type.FullName & " is not marked as serializable.")
        End If
    End Function

    ''' <summary>
    ''' Checks the validity of object names.
    ''' </summary>
    ''' <param name="objectName">An object's name which should be tested for validity.</param>
    ''' <param name="file">Specifies the ini file to use when testing the object name.</param>
    Public Overridable Function IsValidObjectName(ByVal objectName As String, ByVal file As IniFile) As Boolean
        For Each charValue As Char In objectName
            If charValue = chrSeparator OrElse charValue = Chr(10) OrElse charValue = Chr(13) OrElse charValue = file.CommentSeparator OrElse charValue = file.KeyValueSeparator Then
                Return False
            End If
        Next
        Return True
    End Function

    ''' <summary>
    ''' Checks the validity of object names.
    ''' </summary>
    ''' <param name="objectName">An object's name which should be tested for validity.</param>
    ''' <param name="keyValueSeparator">A key-value separator used in ini files.</param>
    ''' <param name="commentSeparator">A comment separator used in ini files.</param>
    Public Overridable Function IsValidObjectName(ByVal objectName As String, ByVal keyValueSeparator As Char, ByVal commentSeparator As Char) As Boolean
        For Each charValue As Char In objectName
            If charValue = chrSeparator OrElse charValue = Chr(10) OrElse charValue = Chr(13) OrElse charValue = commentSeparator OrElse charValue = keyValueSeparator Then
                Return False
            End If
        Next
        Return True
    End Function

    ''' <summary>
    ''' Checks the validity of object names. Uses the default key-value pair and comment separator.
    ''' </summary>
    ''' <param name="objectName">An object's name which should be tested for validity.</param>
    Public Overridable Function IsValidObjectName(ByVal objectName As String) As Boolean
        For Each charValue As Char In objectName
            If charValue = chrSeparator OrElse charValue = Chr(10) OrElse charValue = Chr(13) OrElse charValue = ";"c OrElse charValue = "="c Then
                Return False
            End If
        Next
        Return True
    End Function
End Class
